//  Copyright 2019 root.ext@gmail.com
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import 'dart:ffi';
import 'dart:typed_data';
import 'dart:convert';
import 'package:ffi/ffi.dart';

/// Null-terminated sequence of bytes (UTF-8)
class NativeString extends Struct {
  @Uint8()
  int char;

  static String fromPointer(Pointer pointer) {
    Pointer<Uint8> pointerToBytes = pointer.cast();
    List<int> units = [];
    int len = 0;
    while (true) {
      int char = pointerToBytes.elementAt(len++).value;
      if (char == 0) break;
      units.add(char);
    }
    return Utf8Decoder().convert(units);
  }

  static String fromAddress(int address) {
    return fromPointer(Pointer.fromAddress(address).cast());
  }

  static Pointer<NativeString> fromString(String s) {
    List<int> units = Utf8Encoder().convert(s);
    var result = allocate<Uint8>(count: units.length + 1);
    _copyUint8(units, result);
    result.elementAt(units.length).value = 0;
    return result.cast();
  }
} 

/// Byte buffer
class NativeBuffer extends Struct {
  @Uint8()
  int byte;

  static Pointer<NativeBuffer> fromBytes(List<int> bytes) {
    var result = allocate<Uint8>(count: bytes.length);
    _copyUint8(bytes, result);
    return result.cast();
  }

  static Pointer<NativeBuffer> fromTyped(TypedData data) {
    List<int> bytes = data.buffer.asUint8List();
    return fromBytes(bytes);
  }
}

/// Array of pointers to [NativeString]
class NativeStringArray  {
  Pointer<IntPtr> pointer;
  
  static Pointer<IntPtr> fromList(List<String> strings) {
    var result = allocate<IntPtr>(count: strings.length);
    List<int> pointers = strings.map((s) => NativeString.fromString(s).address).toList();
    _copyIntPtr(pointers, result);
    return result;
  }  
}

void _copyUint8(List<int> bytes, Pointer<Uint8> pointer){
  for (int i = 0; i < bytes.length; i++) {
    pointer.elementAt(i).value = bytes[i];
  }
}

void _copyIntPtr(List<int> bytes, Pointer<IntPtr> pointer){
  for (int i = 0; i < bytes.length; i++) {
    pointer.elementAt(i).value = bytes[i];
  }
}